
//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//	V.R. CONFERENCING APPLICATION                                       //
//                                                                          //
//	Written by R.M.J.Tucker                                             //
//                                                                          //
//	(C) Copyright 1996                                                  //
//                                                                          //
//	VIEW.CPP                                                            //
//                                                                          //
//	Implementation of the VIEWPROJECTION class.                         //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

#include "VIEW\VIEW.H"
#include "TEXTURE\PCX.H"

#include <mem.h>

/*	Only 1024 vertices are catered
	for in the transformation.	*/

#define MAXVERTEX 1024

/*	The global bucket
	contains 1024 vertices ( max ).	*/

POINT vBucket [MAXVERTEX];

/*	The global bucket
	contains 1024 polygons ( max ).	*/

#define MAXPOLYGON 512

/*	Structures internal 
	the view projection.	*/

BUCKETPOLYGON pBucket [MAXPOLYGON];

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Constructor sets up a default camera.                                    //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

VIEWPROJECTION :: VIEWPROJECTION ( void )

/*	PRE	:	TRUE
	POST	:	TRUE
*/

{
	static CAMERA c;

	/*	A default
		camera.	*/

	cWorld = &c;

	/*	and no vertices or
		polygons yet.	*/

	nVertices = nPolygons = 0;

/*	VECTOR	Normal;
	POINT	Point;

	/*	The left clipping
		plane.	*/

	Normal.Set ( -1.0, 0.0, -1.0 );
	Point.Set ( 0.0, 0.0, 1.0 );
	Frustum [0].Set ( Normal, Point );

	/*	The right clipping
		plane.	*/

	Normal.Set ( 1.0, 0.0, -1.0 );
	Point.Set ( 0.0, 0.0, 1.0 );
	Frustum [1].Set ( Normal, Point );

	/*	The top clipping
		plane.	*/

	Normal.Set ( 0.0, 1.0, -1.0 );
	Point.Set ( 0.0, 0.0, 1.0 );
	Frustum [2].Set ( Normal, Point );

	/*	The bottom clipping
		plane.	*/

	Normal.Set ( 0.0, -1.0, -1.0 );
	Point.Set ( 0.0, 0.0, 1.0 );
	Frustum [3].Set ( Normal, Point );

	/*	The back clipping
		plane.	*/

	Normal.Set ( 0.0, 0.0, 1.0 );
	Point.Set ( 0.0, 0.0, 4000.0 );
	Frustum [4].Set ( Normal, Point );

	/*	The front clipping
		plane.	*/

	Normal.Set ( 0.0, 0.0, -1.0 );
	Point.Set ( 0.0, 0.0, 1.0 );
	Frustum [5].Set ( Normal, Point );*/
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Operation to reset the viewing transformation.                           //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

void VIEWPROJECTION :: Reset ( WINDOW& Window )

/*	PRE	:	TRUE
	POST	:	
*/

{
	/*	Empty the
		buckets.	*/

	nVertices = nPolygons = 0;

	/*	and setup the
		view frustum.	*/

	SetupFrustum ( Window );
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Operaton to add the geometry of an object in worldspace to the bucket.   //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

VIEWPROJECTION& VIEWPROJECTION :: operator += ( GEOMETRY& gObject )

/*	PRE	:	TRUE
	POST	:	Projection set = Projection set plus all in `g'
*/
		
{
	/*	add the vertices
		to the vertex bucket.	*/

	movmem ( gObject.Vertices, &vBucket [nVertices], sizeof ( POINT ) * gObject.nVertices );

	/*	add the surfaces to
		the surface bucket.	*/

	REAL	pAngle;
	POLYGON *p;
	int	i;

	for ( i = 0; i < gObject.nPolygons; i++ ) {
		p = &gObject.Polygons [i];

		/*	Determine which side
			of polygon is visible.	*/

		pAngle = cWorld->GetOrientation ( &gObject.Vertices [p->Vertices [0]], &gObject.Vertices [p->Vertices [1]], &gObject.Vertices [p->Vertices [2]] );

		/*	If angle > 0.0 then 
			the outside is visible.	*/

		if ( pAngle > 0 ) {

			/*	Use the outside
				colour.	*/

			pBucket [nPolygons].Colour = p->Outside.Colour;

			/*	And use the outside
				texture.	*/

			pBucket [nPolygons].Texture = p->Outside.Texture;
			pBucket [nPolygons].TexVerts = p->Outside.TexVerts;

			/*	And state that no
				cull occurred.	*/

			pBucket [nPolygons].Culled = FALSE;
		} else {				

			/*	If the inside is visible
				and this surface is to be culled...	*/

			if ( ( p->Flag & CULL ) == CULL ) {
				continue;
			}

			/*	Surface is visible and culled,
				so use inside colour.	*/

			pBucket [nPolygons].Colour = p->Inside.Colour;

			/*	And use the inside
				texture.	*/

			pBucket [nPolygons].Texture = p->Inside.Texture;
			pBucket [nPolygons].TexVerts = p->Inside.TexVerts;

			/*	State that is has
				been culled.	*/

			pBucket [nPolygons].Culled = TRUE;
		}

		/*	Put the polygon into 
			the bucket polygon structure.	*/

		pBucket [nPolygons].p = p;
		pBucket [nPolygons].vBase = nVertices;
		nPolygons++;
	}

	/*	Move the vertex pointer
		along.	*/

	nVertices += gObject.nVertices;

	/*	Return this
		object.	*/

	return ( *this );
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Operaton to apply the global view to the bucket and render it.           //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

void VIEWPROJECTION :: Show ( GRAPHICSMODE& Buffer )

/*	PRE	:	TRUE
	POST	:	All polygons are clipped to the view and rendered
*/

{
	/*	A quick list
		of polygons.	*/

	static struct NODE {
		REAL		Dist;
		BUCKETPOLYGON	*Polygon;
		NODE 		*Next;
	} All_Nodes [MAXPOLYGON];

	/*	A sentinal to ensure
		we always terminate.	*/

	NODE Sentinal = { -100000000.0, NULL, &Sentinal };

	/*	Some list traversal
		nodes.	*/

	NODE *Head = &Sentinal, *Current;

	/*	Apply the global matrix to 
		all of the vertices.	*/

	cWorld->GetMatrix ().Multiply ( nVertices, vBucket, vBucket, cWorld->GetLocation () );

	/*	Now perform an insertion sort on all of the polygons,
		this will be quicker than a `Quick' Sort.	*/

	int	i, nUsed = 0;
	REAL	Dist;

	for ( i = 0; i < nPolygons; i++ ) {

		/*	Calculate the polygon
			distance. */

		Dist = GetDistance ( i );

		/*	Fill the rest of
			the structure.	*/

		All_Nodes [nUsed].Dist = Dist;
		All_Nodes [nUsed].Polygon = &pBucket [i];

		/*	Now insert into
			the list.	*/

		Current = Head;
		while ( Dist < Current->Next->Dist ) {
			Current = Current->Next;
		}

		All_Nodes [nUsed].Next = Current->Next;
		Current->Next = &All_Nodes [nUsed];

		/*	Go to the next
			usable structure.	*/

		nUsed++;
	}

	/*	Tell the mode object which
		window to use.	*/

	Buffer.SetWindow ( Window );

	/*	Now clip the polygons to the 
		view frustum and render them.	*/

	Current = Head->Next;
	while ( Current != &Sentinal ) {
		
		/*	Clip and render 
			the polygon.	*/

		ClipShadeAndRender ( Buffer, Current->Polygon, &vBucket [Current->Polygon->vBase] );

		/*	Go to the
			next polygon.	*/

		Current = Current->Next;
	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Operation to find the centroid of a polygon in the bucket.               //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

REAL VIEWPROJECTION :: GetDistance ( int Index )

/*	PRE	:
	POST	:
*/

{
	int	i, nVerts = pBucket [Index].p->nVerts, Base = pBucket [Index].vBase;
	VECTOR	c ( 0.0, 0.0, 0.0 );

	/*	Weighted sum
		of vertices.	*/

	for ( i = 0; i < nVerts; i++ ) {
		c += ( vBucket [Base + pBucket [Index].p->Vertices [i] ] );
	}

	/*	Find the average
		vertex value.	*/

	c /= nVerts;

	/*	Return the length
		of this vector.	*/

	return ( c.Length () );
}
//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Function sets the camera to use for the viewing transformation.          //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

void VIEWPROJECTION :: SetCamera ( CAMERA *c )

/*	PRE	:
	POST	:
*/

{
	cWorld = c;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Function sets the light source for use by the view transformation.       //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

void VIEWPROJECTION :: SetLight ( CAMERA *c )

/*	PRE	:
	POST	:
*/

{
	cLight = c;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Function gets the camera to use for the viewing transformation.          //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

CAMERA * VIEWPROJECTION :: GetCamera ( void )

/*	PRE	:
	POST	:
*/

{
	return ( cWorld );
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Return the light source camera.                                          //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

CAMERA * VIEWPROJECTION :: GetLightSource ( void )

/*	PRE	:
	POST	:
*/

{
	return ( cLight );
}

WINDOW *VIEWPROJECTION :: GetWindow ( void )

/*	PRE	:
	POST	:
*/

{
	return ( Window );
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// Project a series of points according to the viewing parameters.          //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

void VIEWPROJECTION :: Project ( int n, POINT *In, INTPOINT *Out )

/*	PRE	:
	POST	:
*/

{
	REAL zRecip, x, y;

	while ( --n > -1 ) {

		/*	Find z
			reciprocal.	*/

		zRecip = 1.0 / In [n].v [2];

		/*	Transform the x
			co-ordinate.	*/

		x = In [n].v [0] * zRecip * maxScale + xCenter;
		y = In [n].v [1] * zRecip * maxScale + yCenter;
		Out [n].x = ( int ) x + Window->Left;
		Out [n].y = ( Window->Length- ( int ) y ) + Window->Top;
	}
}




